@@ -2,20 +2,20 @@
 use strict;
 use warnings;
 
-binmode(STDOUT, ":utf8");
-
 use File::Basename;
 use Getopt::Long;
-use Parse::Win32Registry qw(iso8601);
+use Parse::Win32Registry qw(iso8601 hexdump);
+
+binmode(STDOUT, ":utf8");
 
 Getopt::Long::Configure('bundling');
 
 GetOptions('last|l=f'  => \my $period,
-           'recurse|r' => \my $recurse,
-           'values|v'  => \my $show_values);
+           'values|v'  => \my $show_values,
+           'hexdump|x' => \my $show_hexdump);
 
 my $filename = shift or die usage();
-my $initial_key_name = shift;
+my $initial_key_path = shift;
 
 my $registry = Parse::Win32Registry->new($filename)
     or die "'$filename' is not a registry file\n";
@@ -26,10 +26,10 @@ if (!defined($root_key->get_timestamp)) {
     die "'$filename' needs to be an NT-based registry file\n"
 }
 
-if (defined($initial_key_name)) {
-    $root_key = $root_key->get_subkey($initial_key_name);
+if (defined($initial_key_path)) {
+    $root_key = $root_key->get_subkey($initial_key_path);
     if (!defined($root_key)) {
-        die "Could not locate the key '$initial_key_name' in '$filename'\n";
+        die "Could not locate the key '$initial_key_path' in '$filename'\n";
     }
 }
 
@@ -43,7 +43,7 @@ traverse($root_key);
 
 sub traverse {
     my $key = shift;
-    
+
     my $timestamp = $key->get_timestamp;
     push @{$keys_by_timestamp{$timestamp}}, $key;
     $first_timestamp = $timestamp if $timestamp < $first_timestamp;
@@ -64,7 +64,16 @@ foreach my $timestamp (sort keys %keys_by_timestamp) {
         print iso8601($timestamp), "\t", $key->get_path, "\n";
         if ($show_values) {
             foreach my $value ($key->get_list_of_values) {
-                print "\t", $value->as_string, "\n";
+                if (!$show_hexdump) {
+                    print "\t", $value->as_string, "\n";
+                }
+                else {
+                    my $value_name = $value->get_name;
+                    $value_name = "(Default)" if $value_name eq "";
+                    my $value_type = $value->get_type_as_string;
+                    print "\t$value_name ($value_type):\n";
+                    print hexdump($value->get_raw_data);
+                }
             }
             print "\n";
         }
@@ -78,9 +87,10 @@ $script_name for Parse::Win32Registry $Parse::Win32Registry::VERSION
 
 Displays the keys and values of a registry file in date order.
 
-$script_name <filename> [subkey] [-l <number>] [-v]
+$script_name <filename> [subkey] [-l <number>] [-v] [-x]
     -l or --last        display only the last <number> days
                         of registry activity
     -v or --values      display values
+    -x or --hexdump     display value data as a hex dump
 USAGE
 }
